% build.pl
% Author: Silvio do Lago Pereira

:- module(build, [build/3]).
:- use_module(varname).
:- dynamic node/7, cut/2, '$ANSWER'/2.

% built(+Query, -Result, -Tree)

build(Query, Result, Tree) :-
    (   nonvar(Query)
    ->  reset,
        trans(0, Query, Goals),
        namevars(Goals, _),
        free_variables(Goals, Variables),
        maplist([X,Z]>>(get_varname(X,Y), Z = (Y=X)), Variables, Answer),
        append(Goals, ['$ANSWER'(Answer)], NewGoals),
        catch((ignore(sldnf(NewGoals,0,all)), Result = done), Result, true),
        findall(node(N,F,G,M,T,X,Y), retract(node(N,F,G,M,T,X,Y)), Tree),
        retractall(cut(_,_))
    ;   Result = error(instantiation_error,context(_:build/3,_)),
        Tree = [] ).

% Reset flags and facts

reset :-
    reset_varname,
    flag(node, _, 1),
    flag(query, _, 1),
    flag(maxnode, _, 0),
    flag(answer, _, false),
    flag(depth, _, 500),
    flag(positive, _, 1),
    retractall(node(_, _, _, _, _, _, _)),
    retractall(cut(_, _)),
    retractall('$ANSWER'(_, _)).

% Convert conjunction to list, translating cuts

trans(Father, (!,B), [cut(Father)|C]) :- !, trans(Father, B, C).
trans(Father, (A,B), [A|C]) :- !, trans(Father, B, C).
trans(Father, '!', [cut(Father)]) :- !.
trans(_, true, []) :- !.
trans(_, A, [A]) :- !.

% SLDNF proof procedure

sldnf(['$ANSWER'(Answer)], Father, Mode) :- !,
    is_active(Father, Mode),
    new_node(Mode, Node, Father, true, second),
    maplist([X,Y]>>(X=(V=W), namevars(W,N), Y=(V=N)), Answer, NamedAnswer),
    exclude([X]>>(X=(V=V)), NamedAnswer, NamedAnswerSimplified),
    terms_to_atoms(NamedAnswerSimplified, ListOfAtoms),
    assert('$ANSWER'(Node, ListOfAtoms)).

sldnf([Goal|_], Father, Mode) :-
    flag(depth, Limit, Limit),
    node_level(Father, FatherLevel),
    FatherLevel > Limit, !,
    new_node(Mode, _, Father, '...', second),
    throw(error(depth_limit_exceeded, context(Goal))).

sldnf([Goal|_], Father, Mode) :-
    flag(node, Number, Number),
    Number > 20000, !,
    new_node(Mode, _, Father, '...', second),
    throw(error(maximum_number_of_nodes_exceeded, context(Goal))).

sldnf([not(Goal)|Goals], Father, Mode) :- !,
    Query = [not(Goal)|Goals],
    is_active(Father, Mode),
    set_inactive_cut(Father, Query),
    formatted_query(Query, FQuery),
    new_node(Mode, Child, Father, FQuery, second),
    (   not(sldnf([Goal, '$ANSWER'(_)], Child, any))
    ->  last_used_node(NewFather),
        sldnf(Goals, NewFather, Mode)
    ;   last_used_node(NewFather),
        new_node(Mode, _, NewFather, false, second),
        !, fail ).

sldnf([Goal|Goals], Father, Mode) :- !,
    Query = [Goal|Goals],
    is_active(Father, Mode),
    set_inactive_cut(Father, Query),
    formatted_query(Query, FQuery),
    new_node(Mode, Child, Father, FQuery, second),
    (   resolvent(Child, Query, NewQuery, Mgu),
        new_node(Mode, NewFather, Child, Mgu, none),
        flag(answer, _, false),
        sldnf(NewQuery, NewFather, Mode),
        flag(answer, _, true)
    *-> Mode = any, !
    ;   flag(answer, false, false),
        last_used_node(AnotherFather),
        new_node(Mode, _, AnotherFather, false, second),
        fail ).


% Execute predicate cut

resolvent(Father, [cut(Stop)|Goals], Goals, Mgu) :- !,
    formatted_mgu('!', 'builtin', [], Mgu),
    do_cuts(Father, Stop).

% Execute predicate call

resolvent(Father, [call(Goal)|Rest], Goals, Mgu) :- !,
    resolvent(Father, [Goal|Rest], Goals, Mgu).

% Execute predicate (_ -> _ ; _)

resolvent(_, [(Cond->Goal1;Goal2)|Goals], Goals, Mgu) :- !,
    Goal = (Cond->Goal1;Goal2),
    copy_term_nat(Goal, Head),
    user:call(Head),
    unify(Goal, Head, MguList),
    formatted_mgu('->', 'builtin', MguList, Mgu),
    Goal = Head.

% Execute predicate (_ -> _ )

resolvent(_, [(Cond->Goal1)|Goals], Goals, Mgu) :- !,
    Goal = (Cond->Goal1),
    copy_term_nat(Goal, Head),
    user:call(Head),
    unify(Goal, Head, MguList),
    formatted_mgu('->', 'builtin', MguList, Mgu),
    Goal = Head.

% Execute predicate (_ *-> _ ; _)

resolvent(_, [(Cond*->Goal1;Goal2)|Goals], Goals, Mgu) :- !,
    Goal = (Cond*->Goal1;Goal2),
    copy_term_nat(Goal, Head),
    user:call(Head),
    unify(Goal, Head, MguList),
    formatted_mgu('*->', 'builtin', MguList, Mgu),
    Goal = Head.

% Execute predicate (_ *-> _)

resolvent(_, [(Cond*->Goal1)|Goals], Goals, Mgu) :- !,
    Goal = (Cond *-> Goal1),
    copy_term_nat(Goal, Head),
    user:call(Head),
    unify(Goal, Head, MguList),
    formatted_mgu('*->', 'builtin', MguList, Mgu),
    Goal = Head.

% Execute predicate or

resolvent(Father, [(Goal1;Goal2)|Rest], Goals, Mgu) :- !,
    Mgu = '(;.builtin) / {}',
    (   trans(Father, Goal1, List1),
        append(List1, Rest, Goals)
    ;   trans(Father, Goal2, List2),
        append(List2, Rest, Goals)   ).

% Execute predicate and

resolvent(Father, [(!,Goal2)|Rest], Goals, Mgu) :- !,
    G is Father-2,
    assertz(cut(G,off)),
    resolvent(Father, [cut(G),Goal2|Rest], Goals, Mgu).

resolvent(Father,[(Goal1,Goal2)|Rest], Goals, Mgu) :- !,
    resolvent(Father, [Goal1,Goal2|Rest], Goals, Mgu).

% Execute user-defined predicate

resolvent(Father, [Goal|Goals], NewGoal, Mgu) :-
    user_defined(Goal), !,
    nth_clause(user:Goal, ClauseNumber, Reference),
    functor(Goal, Predicate, Arity),
    functor(Head, Predicate, Arity),
    clause(user:Head, Body, Reference),
    unify(Goal, Head, MguList),
    trans(Father, Body, Conditions),
    formatted_mgu(Predicate, ClauseNumber, MguList, Mgu),
    Goal = Head,
    append(Conditions, Goals, NewGoal).

% Execute built-in predicate

resolvent(_, [Goal|Goals], Goals, Mgu) :-
    builtin(Goal), !,
    functor(Goal, Predicate, _),
    copy_term_nat(Goal, Head),
    user:call(Head),
    unify(Goal, Head, MguList),
    formatted_mgu(Predicate, 'builtin', MguList, Mgu),
    Goal = Head.

resolvent(_, [Goal|_], _, _) :-
    throw(error(existence_error(procedure,Goal),_)).


% Checks whether a Goal is a user-defined or built-in

user_defined(Goal) :-
    predicate_property(Goal, imported_from(user)).

builtin(Goal) :-
    ( predicate_property(Goal, built_in)
    ; predicate_property(Goal, imported_from(Module)),
      memberchk(Module, [system, lists, apply, yall, edinburgh, ansi_term]) ),
    !.


% Finds the mgu of two terms (with named varibles)

unify(Goal, Head, MguList) :-
    term_variables(Head, Variables),
    namevars(Variables, _),
    unifiable(Goal, Head, MguList).

% Checks whether a node is still active (not cutted)

is_active(Father, Mode) :-
    grandfather(Father, GrandFather),
    (   not(cut(GrandFather, on))
    ->  true
    ;   new_node(Mode, _Child, Father, cut, second),
        fail ).

set_inactive_cut(Father, Query) :-
    (   intersection([cut(_),!], Query, [])
    ->  true
    ;   grandfather(Father, GrandFather),
        asserta(cut(GrandFather, off)) ).

grandfather(Father, GrandFather) :-
    (   Father = 0
    ->  GrandFather = 0
    ;   node(Father,GrandFather,_,_,_,_,_) ).

% Activate cuts

do_cuts(0, _) :- !.
do_cuts(Start, Stop) :-
    ignore( ( retract(cut(Start, off)),
              assertz(cut(Start, on)) ) ),
    node(Start,Father,_,_,_,_,_),
    (   Start \= Stop
    ->  do_cuts(Father, Stop)
    ;   true ).


% Node creation predicate


new_node(Mode, Node, Father, Goal, Type) :-
    node_level(Father, Level),
    (   node(Father,_,FatherText,_,_,_,_),
        get(FatherText, value, Atom),
        memberchk(Atom, [Goal,cut])
    ->  Node = Father
    ;   flag(node,Node,Node+1),
        assertz(node(Node,Father,Goal,Mode,Type,0,Level)) ).

node_level(Father, Level) :-
    (   Father = 0
    ->  Level is 1
    ;   node(Father, _, _, _, _, _, FatherLevel),
        Level is FatherLevel + 1 ).


last_used_node(F) :-
    flag(node,M,M),
    F is M-1.

% formatted_query(+Number,+Query,-Atom)

formatted_query(Query, Atom) :-
    (  memberchk(Query, [false, true, cut])
    -> format(atom(Atom), '~w', [Query])
    ;  append(Goals, ['$ANSWER'(_)], Query),
       namevars(Goals, NormalizedGoals),
       terms_to_atom(NormalizedGoals, AuxAtom),
       flag(query, Number, Number+1),
       format(atom(Atom),'(~w) ?- ~w.', [Number, AuxAtom]) ), !.


% terms_to_atom(+ListOfTerms,-Atom)

terms_to_atom(A,A) :- atomic(A), !.

terms_to_atom(ListOfTerms,Atom) :-
    terms_to_atoms(ListOfTerms,ListOfAtoms),
    atomic_list_concat(ListOfAtoms,', ',Atom).

% terms_to_atoms(+ListOfTerms,-ListOfAtoms)

terms_to_atoms(ListOfTerms, ListOfAtoms) :-
    maplist([T,A]>>( T = cut(_)  -> format(atom(A), '~w', [!])
                   ; T = call(G) -> format(atom(A), '~w', [G])
                   ; T = (_;_)   -> format(atom(A), '(~w)', [T])
                   ;              format(atom(A), '~w', [T])  ),
            ListOfTerms,
            ListOfAtoms).


% formatted_mgu(+Predicate,+ClauseNumber,+Mgu,-Atom)

formatted_mgu(Predicate,ClauseNumber,Mgu,Atom) :-
    composed_normalized_mgu(Mgu,ComposedNormalizedMgu),
    predsort(compare_pairs,ComposedNormalizedMgu,SortedNormalizedMgu),
    terms_to_atoms(SortedNormalizedMgu,ListOfAtoms),
    atomic_list_concat(ListOfAtoms,', ',AuxAtom),
    format(atom(Atom),'(~w.~w) / {~w}',[Predicate,ClauseNumber,AuxAtom]).

compare_pairs(Relation, V=_, W=_) :-
    varcmp(Relation, V, W).

composed_normalized_mgu(Mgu, ComposedNormalizedMgu) :-
    namevars(Mgu, NormalizedMgu1),
    maplist(call, Mgu),
    namevars(Mgu, NormalizedMgu2),
    maplist([V1=_, _=V2, V1=V2]>>true, NormalizedMgu1, NormalizedMgu2, ComposedNormalizedMgu), !.

composed_normalized_mgu(Mgu, NormalizedMgu) :-
    namevars(Mgu, NormalizedMgu).













