% SLDNF-Tree (version 1.0) - 2018
% Draw SLDNF trees iteratively, with cut and built-in predicates
%
% Author: Silvio do Lago Pereira
% http://www.ime.usp.br/~slago/sldnf.zip

:- module(tree,
          [ tree/0
          , tree/1
          ]).
:- use_module(library(pce)).
:- pce_autoload(finder, library(find_file)).
:- use_module(varname).
:- use_module(build).
:- use_module(layout).
:- pce_global(@finder, new(finder)).
:- dynamic current_window/1, node/4.


tree :-
    tree('').

tree(Goal) :-
    reset_varname,
    flag(maxnode, _, 0),
    retractall(node(_, _, _, _)),
    (   running(W)
    ->  ansi_format([fg(green)], '\nSLDNF-Tree is running at ~w~n', [W]),
        send(W, clear_picture),
        send(W, status_message, 'Restarted!')
    ;   ansi_format([fg(green)], '\nStarting SLDNF-Tree...\n', []),
        new(W, gui('SLDNF-Tree (version 1.0)')),
        send(W, action, restart),
        send(W, action, step),
        send(W, action, go),
        send(W, action, save),
        send(W, action, about),
        send(W, action, quit),
        send(W, open, point(0,0), normalise := @on),
        send(W, status_message, 'Ready!'),
        retractall(current_window(_)),
        assert(current_window(W))  ),
    pretty_atom(Goal, Query),
    send(W, query, Query).


% Actions performed by buttons

restart(W) :-
    send(W, query, ''),
    send(W, clear_picture),
    send(W, status_message, 'Restarted!'),
    reset_varname,
    flag(node, _, 0),
    flag(maxnode, _, 0),
    retractall(node(_,_,_,_)).

step(W) :-
    nextnode(W, N),
    flag(maxnode, M, M),
    (   N < M
    ->  node(N, T, X, Y),
        send(W, draw_node, T, point(X+5, Y+5)),
        ( build:'$ANSWER'(N, A)
        ->  atomic_list_concat(A, ', ', Answer),
            send(W, status_message, string(Answer))
        ;   true )
    ;   send(W, error_message, 'Tree is complete!') ).

go(W) :-
    nextnode(W,_),
    forall(node(_, T, X, Y), send(W, draw_node, T, point(X+5, Y+5))),
    flag(maxnode, M, M),
    flag(node, _, M).

save(W) :-
    get(@finder, file, save, [tuple('JPEG',jpeg), tuple('GIF',gif), tuple('Postscript',ps)], Name),
    send(W, status_message, string('Tree was saved in %s...', Name)),
    get(file(Name), directory_name, Directory),
    cd(Directory),
    file_extension(Name, Extension),
    (   Extension = ps,
        get(W?slot_picture, postscript, Postscript),
        new(F, file(Name)),
        send(F, open, write),
        send(F, append, Postscript),
        send(F, close), !
    ;   (Extension = jpeg ; Extension = gif),
        findall(X, (node(_,T,Xt,_), get(T,width,Wt), X is Xt+Wt), Xs),
        findall(Y, (node(_,T,_,Yt), get(T,height,Ht), Y is Yt+Ht), Ys),
        max_list(Xs, Bw),
        max_list(Ys, Bh),
        new(B, pixmap(@nil, width:=Bw+10, height:=Bh+10)),
        send(W?slot_picture?graphicals, for_all, message(B, draw_in, @arg1)),
        send(B, save, Name, Extension), !
    ;   send(W, error_message, 'Tree was not saved!')).

about(_) :-
   new(D, dialog('About')),
   send(D, append(new(V, view(about, size(37,16))))),
   send(V, font, font(helvetica, normal, 13)),
   send(V, colour, blue),
   send(V, editable, @off),
   send(V, append, 'SLDNF-Tree (version 1.0)\n'),
   send(V, append, 'Copyright  2018 by Silvio do Lago Pereira\n'),
   send(V, append, 'Download: www.ime.usp.br/~slago/sldnf.zip\n\n'),
   send(V, append, 'Deals with:\n'),
   send(V, append, ' negations as failure and cut\n'),
   send(V, append, ' control predicates (-> and *->)\n'),
   send(V, append, ' lists and arithmetic expressions\n'),
   send(V, append, ' built-in and library predicates\n\n'),
   send(V, append, 'Limitations:\n'),
   send(V, append, ' dicts are not supported\n'),
   send(V, append, ' CLP(FD) predicates are not recognized\n'),
   send(V, append, ' predicate repeat/0 generates infinite tree\n'),
   send(V, append, ' atoms with escape sequences are not fully supported\n'),
   send(D, append(button(ok, message(D,return,yes)))),
   get(D, confirm, _),
   send(D, destroy).

quit(W) :-
    send(W, destroy),
    retractall(current_window(_)),
    ansi_format([fg(green)], '\n\nFinishing SLDNF-Tree...\n\n',[]).


% Auxiliary predicates


running(W) :-
    current_window(W),
    object(W).

start(W) :-
    get(W, query, Atom),
    Atom  \=  '',
    send(W, clear_picture),
    catch( read_term_from_atom(Atom, Query, []),
           Error,
           ( error_atom(Error, Message),
             send(W, error_message, Message),
             fail ) ),
    build(Query, Result, Tree),
    error_atom(Result, Message),
    (   Result = done
    ->  send(W, status_message, Message)
    ;   send(W, error_message, Message) ),
    Tree \= [],
    layout(Tree, NewTree),
    retractall(node(_,_,_,_)),
    forall(member(Node,NewTree), assertz(Node)),
    length(Tree, M),
    MaxNode is M+1,
    flag(node, _, 1),
    flag(maxnode, _, MaxNode).

nextnode(W,N) :-
    (   (   flag(maxnode, 0, 0)
        ;   get(W?slot_query, modified, @on) )
    ->  start(W)
    ;   true ),
    flag(node, N, N+1).

file_extension(Name, Extension) :-
    atom_chars(Name, Chars),
    reverse(Chars, Reverse),
    append(Prefix, ['.'|_], Reverse), !,
    reverse(Prefix, NewChars),
    atom_chars(Extension, NewChars).


error_atom(done, 'Tree successfully built!') :- !.

error_atom(error(syntax_error(Cause),string(String,_)),Message) :- !,
    Cause =.. [Culprit|_],
    atom_chars(Culprit,[Start|Rest]),
    upcase_atom(Start,NewStart),
    maplist([X,Y]>>(X='_' -> Y=' ' ; Y=X),[NewStart|Rest],Chars),
    atomic_list_concat(Chars,NewCulprit),
    format(atom(Message),'ERROR: ~w in \"~w\".',[NewCulprit,String]).

error_atom(error(existence_error(procedure,Predicate),_), Message) :- !,
    format(atom(Message), 'ERROR: Undefined predicate ~w.', [Predicate]).

error_atom(error(instantiation_error,context(_:Predicate,_)), Message) :-  !,
    format(atom(Message), 'ERROR: Arguments are not sufficiently instantiated in ~w.', [Predicate]).

error_atom(error(depth_limit_exceeded,context(Goal)), Message) :- !,
    format(atom(Message), 'ERROR: Depth limit exceeded at ~w.', [Goal]).

error_atom(error(resource_error(stack),local), 'Error: Out of local stack!' ) :- !.

error_atom(error(maximum_number_of_nodes_exceeded, context(Goal)), Message) :- !,
    format(atom(Message), 'ERROR: Maximum number of nodes exceeded at ~w.', [Goal]).

error_atom(Error, Message) :-
    writeln(exception:Error),
    message_to_string(Error, String),
    string_chars(String, Chars),
    maplist([X,Y]>>(X='\n' -> Y=' ' ; Y=X), Chars, Rest),
    atomic_list_concat(Rest, Message).



% Graphical user interface class


:- pce_begin_class(gui, frame).

variable(slot_query,   text_item, both).
variable(slot_picture, picture,   get).
variable(slot_report,  label,     get).

initialise(W, Title:name) :->
    send(W, send_super, initialise, Title),
    send(W, done_message, message(W, destroy)),
    send(W, append(new(D, dialog))),
    send(D, append(new(Q, text_item(query)))),
    send(D, append(button(restart))),
    send(D, append(button(step))),
    send(D, append(button(go))),
    send(D, append(button(save))),
    send(D, append(button(about))),
    send(D, append(button(quit))),
    send(D, above(new(P, picture))),
    send(P, above(new(E, dialog))),
    send(E, append(new(R, label))),
    send(Q, value_font(font(screen, normal, 15))),
    send(P, size(size(940,500))),
    send(R, font(font(helvetica, roman, 14))),
    send(W, slot, slot_query, Q),
    send(W, slot, slot_picture, P),
    send(W, slot, slot_report, R).

query(W, Query:name) :->
    "Get query slot value"::
    send(W?slot_query, selection, Query).

query(W, Query:name) :<-
    "Set query slot value"::
    get(W?slot_query, selection, Query).

clear_picture(W) :->
    "Clear picture area"::
    send(W?slot_picture, clear),
    send(W?slot_picture, scroll_to, point(0,0)).

draw_node(W, T:text, P:point) :->
    "Draw node and set focus on it"::
    send(W?slot_picture, display, T, P),
    (   get(T, attribute(boxed), @on)
    ->  send(W?slot_picture, display, box(T?width ,T?height), point(T?x, T?y))
        ;   true ),
    get(T, position, point(Xt,Yt)),
    get(T, width, Wt),
    get(T, height, Ht),
    get(W?slot_picture, visible, area(Xp, Yp, Wp, Hp)),
    (   Xt    < Xp    ->  X is Xt
    ;   Xt+Wt > Xp+Wp ->  X is Xt+Wt-Wp
    ;                     X is Xp),
    (   Yt    < Yp    ->  Y is Yt-50
    ;   Yt+Ht > Yp+Hp ->  Y is Yt+Ht-Hp+50
    ;                     Y is Yp),
    send(W?slot_picture, scroll_to, point(X,Y)).

error_message(W, Message:string) :->
    "Set report slot value"::
    send(W?slot_report, colour, red),
    send(W?slot_report, selection, Message),
    put(7), flush_output.

status_message(W, Message:string) :->
    "Set report slot value"::
    send(W?slot_report, colour, blue),
    send(W?slot_report, selection, Message).

action(W, Predicate:name) :->
    "Set message button with predicate, both with the same name"::
    get(W, member, dialog, D),
    get(D, member, Predicate, Reference),
    send(Reference, message, message(@prolog, Predicate, W)).

:- pce_end_class.
